import logging
from typing import Any, Dict, List, Optional

import requests


default_log = logging.getLogger(__name__)


class ApacheStormClient:
    def __init__(self, url: str, log: logging.Logger = default_log, window: str = ":all-time"):
        self.base_url = f"{url.rstrip('/')}/api/v1"
        self.log = log
        self.window = window

    def make_request(self, url, method="GET", params: Optional[Dict[str, str]] = None):
        self.log.debug(f"Making {method} request to {url}")
        r = requests.request(method, url, timeout=30, verify=False, params=params)
        self.log.debug(f"Received {r} from {url}")
        if r.status_code >= 400:
            raise Exception(f'ApacheStormClient - Got {r} while calling "{url}"')
        else:
            return r.json()

    def cluster_summary(self) -> "ClusterSummary":
        url = f"{self.base_url}/cluster/summary"
        return ClusterSummary(self.make_request(url))

    def supervisor_summary(self) -> List["SupervisorSummary"]:
        url = f"{self.base_url}/supervisor/summary"
        return [SupervisorSummary(supervisor) for supervisor in self.make_request(url).get("supervisors", [])]

    def nimbus_summary(self) -> List["NimbusSummary"]:
        url = f"{self.base_url}/nimbus/summary"
        return [NimbusSummary(nimbus) for nimbus in self.make_request(url).get("nimbuses", [])]

    def topology_summary(self) -> List["TopologySummary"]:
        url = f"{self.base_url}/topology/summary"
        return [TopologySummary(topology) for topology in self.make_request(url).get("topologies", [])]

    def topology(self, topology_id) -> "Topology":
        url = f"{self.base_url}/topology/{topology_id}"
        return Topology(self.make_request(url, params={"window": self.window}))


class ClusterSummary:
    def __init__(self, raw_json: Dict[str, Any]):
        self.total_mem = raw_json.get("totalMem")
        self.slots_free = raw_json.get("slotsFree")
        self.mem_assigned_percent_util = raw_json.get("memAssignedPercentUtil")
        self.user = raw_json.get("user")
        self.topologies = raw_json.get("topologies")
        self.avail_mem = raw_json.get("availMem")
        self.total_cpu = raw_json.get("totalCpu")
        self.supervisors = raw_json.get("supervisors")
        self.central_log_url = raw_json.get("central-log-url")
        self.cpu_assigned_percent_util = raw_json.get("cpuAssignedPercentUtil")
        self.bugtracker_url = raw_json.get("bugtracker-url")
        self.tasks_total = raw_json.get("tasksTotal")
        self.avail_cpu = raw_json.get("availCpu")
        self.slots_used = raw_json.get("slotsUsed")
        self.scheduler_display_resource = raw_json.get("schedulerDisplayResource")
        self.slots_total = raw_json.get("slotsTotal")
        self.storm_version = raw_json.get("stormVersion")
        self.executors_total = raw_json.get("executorsTotal")


# self.$1 = raw_json.get("$1")
class SupervisorSummary:
    def __init__(self, raw_json: Dict[str, Any]):

        self.avail_cpu: float = raw_json.get("availCpu")
        self.avail_mem: float = raw_json.get("availMem")
        self.host: str = raw_json.get("host")
        self.id: str = raw_json.get("id")
        self.log_link: str = raw_json.get("logLink")
        self.slots_free: float = raw_json.get("slotsFree")
        self.slots_total: float = raw_json.get("slotsTotal")
        self.slots_used: float = raw_json.get("slotsUsed")
        self.total_cpu: float = raw_json.get("totalCpu")
        self.total_mem: float = raw_json.get("totalMem")
        self.uptime: str = raw_json.get("uptime")
        self.uptime_seconds: float = raw_json.get("uptimeSeconds")
        self.used_cpu: float = raw_json.get("usedCpu")
        self.used_mem: float = raw_json.get("usedMem")
        self.version: str = raw_json.get("version")


class NimbusSummary:
    def __init__(self, raw_json: Dict[str, Any]):
        self.host: str = raw_json.get("host")
        self.port: float = raw_json.get("port")
        self.nimbus_log_link: str = raw_json.get("nimbusLogLink")
        self.status: str = raw_json.get("status")
        self.version: str = raw_json.get("version")
        self.nimbus_up_time: str = raw_json.get("nimbusUpTime")
        self.nimbus_up_time_seconds: float = raw_json.get("nimbusUpTimeSeconds")


class TopologySummary:
    def __init__(self, raw_json: Dict[str, Any]):
        self.assigned_cpu: float = raw_json.get("assignedCpu")
        self.assigned_mem_off_heap: float = raw_json.get("assignedMemOffHeap")
        self.assigned_mem_on_heap: float = raw_json.get("assignedMemOnHeap")
        self.assigned_total_mem: float = raw_json.get("assignedTotalMem")
        self.encoded_id: str = raw_json.get("encodedId")
        self.executors_total: float = raw_json.get("executorsTotal")
        self.id: str = raw_json.get("id")
        self.name: str = raw_json.get("name")
        self.owner: str = raw_json.get("owner")
        self.replication_count: float = raw_json.get("replicationCount")
        self.requested_cpu: float = raw_json.get("requestedCpu")
        self.requested_mem_off_heap: float = raw_json.get("requestedMemOffHeap")
        self.requested_mem_on_heap: float = raw_json.get("requestedMemOnHeap")
        self.requested_total_mem: float = raw_json.get("requestedTotalMem")
        self.scheduler_info: str = raw_json.get("schedulerInfo")
        self.status: str = raw_json.get("status")
        self.tasks_total: float = raw_json.get("tasksTotal")
        self.uptime: str = raw_json.get("uptime")
        self.uptime_seconds: float = raw_json.get("uptimeSeconds")
        self.workers_total: float = raw_json.get("workersTotal")


class Topology(TopologySummary):
    def __init__(self, raw_json: Dict[str, Any]):
        super().__init__(raw_json)
        self.spouts: List["Spout"] = [Spout(spout) for spout in raw_json.get("spouts", [])]
        self.bolts: List["Bolt"] = [Bolt(bolt) for bolt in raw_json.get("bolts", [])]


class Spout:
    def __init__(self, raw_json: Dict[str, Any]):
        self.acked: float = raw_json.get("acked")
        self.complete_latency: str = raw_json.get("completeLatency")
        self.emitted: float = raw_json.get("emitted")
        self.encoded_spout_id: str = raw_json.get("encodedSpoutId")
        self.error_host: str = raw_json.get("errorHost")
        self.error_lapsed_secs: str = raw_json.get("errorLapsedSecs")
        self.error_port: str = raw_json.get("errorPort")
        self.error_time: str = raw_json.get("errorTime")
        self.error_worker_log_link: str = raw_json.get("errorWorkerLogLink")
        self.executors: float = raw_json.get("executors")
        self.failed: float = raw_json.get("failed")
        self.last_error: str = raw_json.get("lastError")
        self.requested_cpu: float = raw_json.get("requestedCpu")
        self.requested_mem_off_heap: float = raw_json.get("requestedMemOffHeap")
        self.requested_mem_on_heap: float = raw_json.get("requestedMemOnHeap")
        self.spout_id: str = raw_json.get("spoutId")
        self.tasks: float = raw_json.get("tasks")
        self.transferred: float = raw_json.get("transferred")


class Bolt:
    def __init__(self, raw_json: Dict[str, Any]):
        self.acked: float = raw_json.get("acked")
        self.bolt_id: str = raw_json.get("boltId")
        self.capacity: str = raw_json.get("capacity")
        self.emitted: float = raw_json.get("emitted")
        self.encoded_bolt_id: str = raw_json.get("encodedBoltId")
        self.error_host: str = raw_json.get("errorHost")
        self.error_lapsed_secs: str = raw_json.get("errorLapsedSecs")
        self.error_port: str = raw_json.get("errorPort")
        self.error_time: str = raw_json.get("errorTime")
        self.error_worker_log_link: str = raw_json.get("errorWorkerLogLink")
        self.execute_latency: str = raw_json.get("executeLatency")
        self.executed: float = raw_json.get("executed")
        self.executors: float = raw_json.get("executors")
        self.failed: float = raw_json.get("failed")
        self.last_error: str = raw_json.get("lastError")
        self.process_latency: str = raw_json.get("processLatency")
        self.requested_cpu: float = raw_json.get("requestedCpu")
        self.requested_mem_off_heap: float = raw_json.get("requestedMemOffHeap")
        self.requested_mem_on_heap: float = raw_json.get("requestedMemOnHeap")
        self.tasks: float = raw_json.get("tasks")
        self.transferred: float = raw_json.get("transferred")
