from concurrent.futures import ThreadPoolExecutor
import time

from ruxit.api.base_plugin import RemoteBasePlugin
from ruxit.api.topology_builder import Group, Device

from apache_storm_api import ApacheStormClient, TopologySummary


class ApacheStormExtension(RemoteBasePlugin):
    def initialize(self, **kwargs):
        self.executions = -1
        self.logger.setLevel(self.config.get("log_level", "INFO"))

    def query(self, **kwargs):

        frequency = self.config.get("frequency", 1)
        self.frequency_seconds = frequency * 60
        self.executions += 1

        if self.executions % frequency == 0:
            start_time = time.time()
            cluster_name = self.config.get("storm_cluster_name")
            self.logger.info(f"Starting query for {cluster_name}")

            self.supervisor_group: Group = self.topology_builder.create_group(f"Supervisors ({cluster_name})", f"Supervisors ({cluster_name})")
            self.nimbus_group: Group = self.topology_builder.create_group(f"Nimbus ({cluster_name})", f"Nimbus ({cluster_name})")
            self.topology_group: Group = self.topology_builder.create_group(f"Topology ({cluster_name})", f"Topology ({cluster_name})")

            self.client: ApacheStormClient = ApacheStormClient(self.config.get("storm_url"), log=self.logger, window=f"{self.frequency_seconds}")

            with ThreadPoolExecutor(max_workers=10) as e:
                e.submit(self.process_cluster)
                e.submit(self.process_supervisor)
                e.submit(self.process_nimbus)
                e.submit(self.process_topologies)
            execution_time = time.time() - start_time
            self.logger.info(f"Finished query for {cluster_name} in {execution_time:.2f}s")
            self.topology_group.absolute("execution_time", execution_time)
        else:
            self.logger.info(f"Not running, next execution in {frequency - (self.executions % frequency)} minutes.")

    def process_cluster(self):
        try:
            cluster_summary = self.client.cluster_summary()

            self.topology_group.report_property("Storm version", cluster_summary.storm_version)
            self.topology_group.report_property("Total memory", f"{cluster_summary.total_mem}")
            self.topology_group.report_property("Total CPU", f"{cluster_summary.total_cpu}")
            self.topology_group.report_property("Total worker slots", f"{cluster_summary.slots_total}")
            self.topology_group.report_property("Tasks", f"{cluster_summary.tasks_total}")

            self.topology_group.absolute("cluster_memory_used", cluster_summary.mem_assigned_percent_util)
            self.topology_group.absolute("cluster_cpu_used", cluster_summary.cpu_assigned_percent_util)
        except Exception as e:
            self.logger.exception(f"Error during process_cluster: {e}")

    def process_supervisor(self):
        try:
            for supervisor in self.client.supervisor_summary():
                self.logger.info(f"Processing supervisor {supervisor.host}...")
                device = self.supervisor_group.create_device(supervisor.id, f"Supervisor {supervisor.host}")
                device.report_property("Version", supervisor.version)
                device.report_property("ID", supervisor.id)
                device.report_property("Log link", supervisor.log_link)
                device.report_property("Total CPU", f"{supervisor.total_cpu}")
                device.report_property("Total Memory", f"{supervisor.total_mem}")

                device.absolute("supervisor_cpu_used", supervisor.used_cpu / supervisor.total_cpu * 100)
                device.absolute("supervisor_memory_used", supervisor.used_mem / supervisor.total_mem * 100)
                device.absolute("supervisor_uptime", supervisor.uptime_seconds)
                device.absolute("supervisor_slots_free", supervisor.slots_free)
                device.absolute("supervisor_slots_used", supervisor.slots_used)
        except Exception as e:
            self.logger.exception(f"Error during process_supervisor: {e}")

    def process_nimbus(self):
        try:
            for nimbus in self.client.nimbus_summary():
                self.logger.info(f"Processing nimbus {nimbus.host}...")
                device = self.nimbus_group.create_device(f"Nimbus {nimbus.host}", f"Nimbus {nimbus.host}")
                device.report_property("Version", nimbus.version)
                device.report_property("Log link", nimbus.nimbus_log_link)
                device.report_property("Port", f"{nimbus.port}")
                device.report_property("Status", nimbus.status)

                device.absolute("nimbus_uptime", nimbus.nimbus_up_time_seconds)
                device.state_metric("nimbus_status", nimbus.status.upper().replace(" ", "_"))
        except Exception as e:
            self.logger.exception(f"Error during process_nimbus: {e}")

    def process_topologies(self):
        try:
            with ThreadPoolExecutor(max_workers=10) as e:
                for topology in self.client.topology_summary():
                    self.logger.info(f"Processing topology summary {topology.name}...")
                    device = self.topology_group.create_device(topology.id, f"Topology {topology.name}")
                    device.report_property("ID", topology.id)
                    device.report_property("Owner", topology.owner)
                    device.report_property("Status", topology.status)

                    device.state_metric("topology_status", topology.status)
                    device.absolute("topology_replication_count", topology.replication_count)
                    device.absolute("topology_tasks", topology.tasks_total)
                    device.absolute("topology_uptime", topology.uptime_seconds)
                    device.absolute("topology_workers", topology.workers_total)
                    device.absolute("topology_assigned_cpu", topology.assigned_cpu)
                    device.absolute("topology_assigned_memory", topology.assigned_total_mem)

                    e.submit(self.process_topology, device, topology)

        except Exception as e:
            self.logger.exception(f"Error during process_topologies: {e}")

    def process_topology(self, device: Device, topology: TopologySummary):
        try:
            self.logger.info(f"Processing topology details for {topology.name}...")
            topology = self.client.topology(topology.id)
            for spout in topology.spouts:
                device.absolute("spout_acked", spout.acked, dimensions={"Spout": spout.spout_id})
                device.absolute("spout_complete_latency", float(spout.complete_latency), dimensions={"Spout": spout.spout_id})
                device.absolute("spout_emitted", spout.emitted, dimensions={"Spout": spout.spout_id})
                device.absolute("spout_failed", spout.failed, dimensions={"Spout": spout.spout_id})
                device.absolute("spout_transferred", spout.transferred, dimensions={"Spout": spout.spout_id})

            for bolt in topology.bolts:
                device.absolute("bolt_acked", bolt.acked, dimensions={"Bolt": bolt.bolt_id})
                device.absolute("bolt_capacity", float(bolt.capacity), dimensions={"Bolt": bolt.bolt_id})
                device.absolute("bolt_emitted", bolt.emitted, dimensions={"Bolt": bolt.bolt_id})
                device.absolute("bolt_execute_latency", float(bolt.execute_latency), dimensions={"Bolt": bolt.bolt_id})
                device.absolute("bolt_failed", bolt.failed, dimensions={"Bolt": bolt.bolt_id})
                device.absolute("bolt_process_latency", float(bolt.process_latency), dimensions={"Bolt": bolt.bolt_id})
                device.absolute("bolt_transferred", bolt.transferred, dimensions={"Bolt": bolt.bolt_id})

        except Exception as e:
            self.logger.exception(f"Error during process_topology({topology.name}): {e}")
